"""VAST Database bucket.

VAST S3 buckets can be used to create Database schemas and tables.
It is possible to list and access VAST snapshots generated over a bucket.
"""

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Optional

from . import errors, schema, transaction

if TYPE_CHECKING:
    from .schema import Schema

log = logging.getLogger(__name__)


@dataclass
class Bucket:
    """VAST bucket."""

    name: str
    tx: "transaction.Transaction"

    def create_schema(self, path: str, fail_if_exists=True) -> "Schema":
        """Create a new schema (a container of tables) under this bucket."""
        if current := self.schema(path, fail_if_missing=False):
            if fail_if_exists:
                raise errors.SchemaExists(self.name, path)
            else:
                return current
        self.tx._rpc.api.create_schema(self.name, path, txid=self.tx.txid)
        log.info("Created schema: %s", path)
        return self.schema(path)  # type: ignore[return-value]

    def schema(self, path: str, fail_if_missing=True) -> Optional["Schema"]:
        """Get a specific schema (a container of tables) under this bucket."""
        s = self.schemas(path)
        log.debug("schema: %s", s)
        if not s:
            if fail_if_missing:
                raise errors.MissingSchema(self.name, path)
            else:
                return None
        assert len(s) == 1, f"Expected to receive only a single schema, but got: {len(s)}. ({s})"
        log.debug("Found schema: %s", s[0].name)
        return s[0]

    def schemas(self, name: Optional[str] = None) -> List["Schema"]:
        """List bucket's schemas."""
        schemas = []
        next_key = 0
        exact_match = bool(name)
        log.debug("list schemas param: schema=%s, exact_match=%s", name, exact_match)
        while True:
            _bucket_name, curr_schemas, next_key, is_truncated, _ = \
                self.tx._rpc.api.list_schemas(bucket=self.name, next_key=next_key, txid=self.tx.txid,
                                               name_prefix=name, exact_match=exact_match)
            if not curr_schemas:
                break
            schemas.extend(curr_schemas)
            if not is_truncated:
                break

        return [schema.Schema(name=name, bucket=self) for name, *_ in schemas]

    def snapshot(self, name, fail_if_missing=True) -> Optional["Bucket"]:
        """Get snapshot by name (if exists)."""
        snapshots, _is_truncated, _next_key = \
            self.tx._rpc.api.list_snapshots(bucket=self.name, name_prefix=name, max_keys=1)

        expected_name = f".snapshot/{name}"
        exists = snapshots and snapshots[0] == expected_name + "/"
        if not exists:
            if fail_if_missing:
                raise errors.MissingSnapshot(self.name, expected_name)
            else:
                return None

        return Bucket(name=f'{self.name}/{expected_name}', tx=self.tx)

    def snapshots(self) -> List["Bucket"]:
        """List bucket's snapshots."""
        snapshots = []
        next_key = 0
        while True:
            curr_snapshots, is_truncated, next_key = \
                self.tx._rpc.api.list_snapshots(bucket=self.name, next_token=next_key)
            if not curr_snapshots:
                break
            snapshots.extend(curr_snapshots)
            if not is_truncated:
                break

        return [
            Bucket(name=f'{self.name}/{snapshot.strip("/")}', tx=self.tx)
            for snapshot in snapshots
        ]



import pytest

from .. import errors


def test_schemas(session, clean_bucket_name):
    with session.transaction() as tx:
        b = tx.bucket(clean_bucket_name)
        assert b.schemas() == []

        s = b.create_schema('s1')
        assert s.bucket == b
        assert b.schemas() == [s]

        s.rename('s2')
        assert s.bucket == b
        assert s.name == 's2'
        assert b.schemas()[0].name == 's2'

        s.drop()
        assert b.schemas() == []


def test_exists(session, clean_bucket_name):
    with session.transaction() as tx:
        b = tx.bucket(clean_bucket_name)
        assert b.schemas() == []

        s = b.create_schema('s1')

        assert b.schemas() == [s]
        with pytest.raises(errors.SchemaExists):
            b.create_schema('s1')

        assert b.schemas() == [s]
        assert b.create_schema('s1', fail_if_exists=False) == s
        assert b.schemas() == [s]


def test_commits_and_rollbacks(session, clean_bucket_name):
    with session.transaction() as tx:
        b = tx.bucket(clean_bucket_name)
        assert b.schemas() == []
        b.create_schema("s3")
        assert b.schemas() != []
        # implicit commit

    with pytest.raises(ZeroDivisionError):
        with session.transaction() as tx:
            b = tx.bucket(clean_bucket_name)
            b.schema("s3").drop()
            assert b.schemas() == []
            1 / 0  # rollback schema dropping

    with session.transaction() as tx:
        b = tx.bucket(clean_bucket_name)
        assert b.schemas() != []


def test_list_snapshots(session, clean_bucket_name):
    with session.transaction() as tx:
        b = tx.bucket(clean_bucket_name)
        b.snapshots()  # VAST Catalog may create some snapshots



import os
from pathlib import Path

import boto3
import pytest

import vastdb


def pytest_addoption(parser):
    parser.addoption("--tabular-bucket-name", help="Name of the S3 bucket with Tabular enabled", default="vastdb")
    parser.addoption("--tabular-access-key", help="Access key with Tabular permissions (AWS_ACCESS_KEY_ID)", default=os.environ.get("AWS_ACCESS_KEY_ID", None))
    parser.addoption("--tabular-secret-key", help="Secret key with Tabular permissions (AWS_SECRET_ACCESS_KEY)", default=os.environ.get("AWS_SECRET_ACCESS_KEY", None))
    parser.addoption("--tabular-endpoint-url", help="Tabular server endpoint", default="http://localhost:9090")
    parser.addoption("--data-path", help="Data files location", default=None)
    parser.addoption("--crater-path", help="Save benchmark results in a dedicated location", default=None)


@pytest.fixture(scope="session")
def session(request):
    return vastdb.connect(
        access=request.config.getoption("--tabular-access-key"),
        secret=request.config.getoption("--tabular-secret-key"),
        endpoint=request.config.getoption("--tabular-endpoint-url"),
    )


@pytest.fixture(scope="session")
def test_bucket_name(request):
    return request.config.getoption("--tabular-bucket-name")


@pytest.fixture(scope="function")
def clean_bucket_name(request, test_bucket_name, session):
    with session.transaction() as tx:
        b = tx.bucket(test_bucket_name)
        for s in b.schemas():
            for t in s.tables():
                t.drop()
            s.drop()
    return test_bucket_name


@pytest.fixture(scope="session")
def s3(request):
    return boto3.client(
        's3',
        aws_access_key_id=request.config.getoption("--tabular-access-key"),
        aws_secret_access_key=request.config.getoption("--tabular-secret-key"),
        endpoint_url=request.config.getoption("--tabular-endpoint-url"))


@pytest.fixture(scope="function")
def parquets_path(request):
    return Path(request.config.getoption("--data-path"))


@pytest.fixture(scope="function")
def crater_path(request):
    return request.config.getoption("--crater-path")



import contextlib
import logging
import threading
from http.server import BaseHTTPRequestHandler, HTTPServer
from itertools import cycle

import pytest
import requests

import vastdb

log = logging.getLogger(__name__)


def test_hello_world(session):
    with session.transaction() as tx:
        assert tx.txid is not None


def test_bad_credentials(session):
    bad_session = vastdb.connect(access='BAD', secret='BAD', endpoint=session.api.url)
    with pytest.raises(vastdb.errors.Forbidden):
        with bad_session.transaction():
            pass


def test_bad_endpoint(session):
    with pytest.raises(requests.exceptions.ConnectionError):
        vastdb.connect(access='BAD', secret='BAD', endpoint='http://invalid-host-name-for-tests:12345')


def test_version_extraction():
    # A list of version and expected version parsed by API
    TEST_CASES = [
            (None, None),                                   # vast server without version in header
            ("5", None),                                    # major
            ("5.2", None),                                  # major.minor
            ("5.2.0", None),                                # major.minor.patch
            ("5.2.0.10", "5.2.0.10"),                       # major.minor.patch.protocol
            ("5.2.0.10 some other things", None),           # suffix
            ("5.2.0.10.20", None),                          # extra version
    ]

    # Mock OPTIONS handle that cycles through the test cases response
    class MockOptionsHandler(BaseHTTPRequestHandler):
        versions_iterator = cycle(TEST_CASES)

        def __init__(self, *args) -> None:
            super().__init__(*args)

        def do_OPTIONS(self):
            self.send_response(204)
            self.end_headers()

        def version_string(self):
            version = next(self.versions_iterator)[0]
            return f"vast {version}" if version else "vast"

        def log_message(self, format, *args):
            log.debug(format, *args)

    # start the server on localhost on some available port port
    server_address = ('localhost', 0)
    httpd = HTTPServer(server_address, MockOptionsHandler)

    def start_http_server_in_thread():
        log.info(f"Mock HTTP server is running on port {httpd.server_port}")
        httpd.serve_forever()
        log.info("Mock HTTP server killed")

    # start the server in a thread so we have the main thread to operate the API
    server_thread = threading.Thread(target=start_http_server_in_thread)
    server_thread.start()

    try:
        for _, expected in TEST_CASES:
            with (pytest.raises(NotImplementedError) if expected is None else contextlib.nullcontext()):
                s = vastdb.connect(endpoint=f"http://localhost:{httpd.server_port}", access="abc", secret="abc")
                assert s.api.vast_version == expected
    finally:
        # make sure we shut the server down no matter what
        httpd.shutdown()



"""VAST database session.

It should be used to interact with a specific VAST cluster.
For more details see:
- [Virtual IP pool configured with DNS service](https://support.vastdata.com/s/topic/0TOV40000000FThOAM/configuring-network-access-v50)
- [S3 access & secret keys on VAST cluster](https://support.vastdata.com/s/article/UUID-4d2e7e23-b2fb-7900-d98f-96c31a499626)
- [Tabular identity policy with the proper permissions](https://support.vastdata.com/s/article/UUID-14322b60-d6a2-89ac-3df0-3dfbb6974182)
"""

import os

import boto3

from . import errors, internal_commands, transaction


class Features:
    """VAST database features - check if server is already support a feature."""

    def __init__(self, vast_version):
        """Save the server version."""
        self.vast_version = vast_version

    def check_imports_table(self):
        """Check if the feature that support imports table is supported."""
        if self.vast_version < (5, 2):
            raise errors.NotSupportedVersion("import_table requires 5.2+", self.vast_version)

    def check_return_row_ids(self):
        """Check if insert/update/delete can return the row_ids."""
        if self.vast_version < (5, 1):
            raise errors.NotSupportedVersion("return_row_ids requires 5.1+", self.vast_version)


class Session:
    """VAST database session."""

    def __init__(self, access=None, secret=None, endpoint=None, ssl_verify=True):
        """Connect to a VAST Database endpoint, using specified credentials."""
        if access is None:
            access = os.environ['AWS_ACCESS_KEY_ID']
        if secret is None:
            secret = os.environ['AWS_SECRET_ACCESS_KEY']
        if endpoint is None:
            endpoint = os.environ['AWS_S3_ENDPOINT_URL']

        self.api = internal_commands.VastdbApi(endpoint, access, secret, ssl_verify=ssl_verify)
        version_tuple = tuple(int(part) for part in self.api.vast_version.split('.'))
        self.features = Features(version_tuple)
        self.s3 = boto3.client('s3',
            aws_access_key_id=access,
            aws_secret_access_key=secret,
            endpoint_url=endpoint)

    def __repr__(self):
        """Don't show the secret key."""
        return f'{self.__class__.__name__}(endpoint={self.api.url}, access={self.api.access_key})'

    def transaction(self):
        """Create a non-initialized transaction object.

        It should be used as a context manager:

            with session.transaction() as tx:
                tx.bucket("bucket").create_schema("schema")
        """
        return transaction.Transaction(self)



import logging

import pyarrow as pa

log = logging.getLogger(__name__)


def test_basic_projections(session, clean_bucket_name):
    with session.transaction() as tx:
        s = tx.bucket(clean_bucket_name).create_schema('s1')
        columns = pa.schema([
            ('a', pa.int8()),
            ('b', pa.int16()),
            ('c', pa.string()),
            ('d', pa.int16()),
            ('e', pa.int64()),
            ('s', pa.struct([('x', pa.int8()), ('y', pa.int16())]))
        ])

        assert s.tables() == []
        t = s.create_table('t1', columns)
        assert s.tables() == [t]

        sorted_columns = ['a']
        unsorted_columns = ['b']
        p1 = t.create_projection('p1', sorted_columns, unsorted_columns)

        sorted_columns = ['b']
        unsorted_columns = ['c', 'd']
        p2 = t.create_projection('p2', sorted_columns, unsorted_columns)

        projs = t.projections()
        assert projs == [t.projection('p1'), t.projection('p2')]
        p1 = t.projection('p1')
        assert p1.name == 'p1'
        p2 = t.projection('p2')
        assert p2.name == 'p2'

        p1.rename('p_new')
        p2.drop()
        projs = t.projections()
        assert len(projs) == 1
        assert projs[0].name == 'p_new'
